package edu.georgetown.gucs.dictionary;

import java.io.Serializable;

import edu.georgetown.gucs.utility.Token;

/**
 * Statistics kept on a per-token basis within a {@linkplain Dictionary}
 *
 * @author Clay Shields
 **/
public class DictionaryEntry implements Serializable {

    /** arbitrary serialVersionUID */
    private static final long serialVersionUID = -279122756982481856L;
    private final Token token;
    private double IDF;
    private int position;
    private int frequencyCount;
    private int documentCount;

    /**
     * Constructor that specifies a token
     *
     * @param token
     *          the token value
     */
    DictionaryEntry(Token token) {
        this.token = token;
        this.frequencyCount = 1;
        this.documentCount = 1;
    }

    /**
     * Constructor that specifies a token and initializes its position, frequency count, document count, and IDF
     *
     * @param token
     *          the token value
     * @param position
     *          the <code>int</code> value of the position in which this token appears
     * @param frequencyCount
     *          the <code>int</code> value of the number of times this token has been seen
     * @param documentCount
     *          the <code>int</code> value of the number of documents this token appears in
     * @param IDF
     *          the <code>double</code> value of the inverse document frequency (IDF) of this token
     **/
    DictionaryEntry(Token token, int position, int frequencyCount, int documentCount, double IDF) {
        this.token = token;
        this.position = position;
        this.frequencyCount = frequencyCount;
        this.documentCount = documentCount;
        this.IDF = IDF;
    }

    /**
     * Provides the position of the token in this dictionary. The position is the most recent location this token has
     * been seen.
     *
     * @return the <code>int</code> value of this token's most recent position
     **/
    public int getPosition() {
        return this.position;
    }

    /**
     * Sets the position of the token in this dictionary. The position is the most recent location this token has been
     * seen.
     *
     * @param position
     *          the <code>int</code> value of the new position
     **/
    public void setPosition(int position) {
        this.position = position;
    }

    /**
     * Provides the token
     *
     * @return the token
     **/
    public Token getToken() {
        return this.token;
    }

    /**
     * Increases the number of times this token has been seen by 1
     **/
    void incrementFrequencyCount() {
        this.frequencyCount++;
    }

    /**
     * Increases the number of times this token has been seen by <code>amount</code>
     *
     * @param amount
     *          the <code>int</code> value to increment the frequency count by
     **/
    void incrementFrequencyCount(int amount) {
        this.frequencyCount += amount;
    }

    /**
     * Provides the frequency count for this entry
     *
     * @return the <code>int</code> value of the number of times this token has been seen
     **/
    public int getFrequencyCount() {
        return this.frequencyCount;
    }

    /**
     * Increases the number of documents this token has been seen in by 1
     **/
    void incrementDocumentCount() {
        this.documentCount++;
    }

    /**
     * Provides the number of documents this token appears in
     *
     * @return the <code>int</code> value of the number of documents this token has been seen in
     **/
    public int getDocumentCount() {
        return this.documentCount;
    }

    /**
     * Provides the IDF of this token
     *
     * @return the <code>double</code> value of the inverse document frequency (IDF) of this token
     **/
    public double getIDF() {
        return this.IDF;
    }

    /**
     * Computes the IDF of this token using the <code>totalDocuments</code>
     *
     * @param totalDocuments
     *          the <code>int</code> value of the total number of documents in the dictionary
     *
     * @return the <code>double</code> value of the inverse document frequency (IDF) of this token
     */
    double computeIDF(int totalDocuments) {
        this.IDF = Math.log((double)totalDocuments/this.documentCount);
        return this.IDF;
    }
}
