package edu.georgetown.gucs.fingerprinter;

import edu.georgetown.gucs.utility.*;

import java.io.File;
import java.math.BigInteger;
import java.util.*;

import edu.georgetown.gucs.dictionary.Dictionary;
import org.slf4j.LoggerFactory;


/**
 * This is a class that allows threaded parallel creation of fingerprints.
 *
 * @author Clay Shields
 */


public class BulkFingerprinter {

    private List<Fingerprint> fpList;
    private int finishedThreadCount = 0 ;
    private final ArrayList<Fingerprint> fingerprint = new ArrayList<>();

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(BulkFingerprinter.class);


    /**
     * A constructor that will set the number of threads equal to the number of cores.
     *
     * @param fingerprinterName the name of the fingerprinter-derived class to use for fingerprinting
     * @param files             A FileLister object instantiated with the directory containing the files to fingerprint
     * @param dictionary        The dictionary to be used for fingerprinting
     */


    public BulkFingerprinter(String fingerprinterName, FileLister files, Dictionary dictionary) throws SdTextException {
        fpList = new ArrayList<Fingerprint>();
    }

    protected Dictionary dict;
    protected Fingerprinter fingerprinter;
    protected String fingerprinterName;

    public BulkFingerprinter(Dictionary dict, String fingerprinterName) throws SdTextException {
        fpList = new ArrayList<>();
        try {

            this.dict = dict;
            fingerprinter = (Fingerprinter) ObjectBuilder.buildObject(fingerprinterName);
            fingerprinter.setDictionary(dict);
            this.fingerprinterName = fingerprinterName;

        } catch (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
    }

    /**
     * Recursively computes all the fingerprints of documents in a given directory as
     * Base64 encoded strings
     *
     * @param fingerprinterName The name of the Fingerprinter class to use
     * @param dictionary The Dictionary object to use for fingerprints
     * @param base the directory to fingerprint
     * @return the Base64 encoded string containing the fingerprint and if applicable the start and end bytes for a split
     * file
     */
    public List<Fingerprint> computeAllFingerprints(String fingerprinterName, Dictionary dictionary, FileLister base) throws SdTextException {

	fpList = Global.pool.invoke(new BulkFingerprinterWorker(fingerprinterName,
							   dictionary,
							   base.getFileList(),
							   0,
							   base.getFileList().size() - 1));
	if (fpList.size() != base.size()){
	    throw new SdTextException("BulkFingerprinter:computeAllFingerprintsUnthreaded:Mismatch in file list and fp list size.");
	}	

	return fpList;		       
    }


    public List<Fingerprint> computeAllFingerprintsUnthreaded(String fingerprinterName, Dictionary dictionary, FileLister base) throws SdTextException {

	
	try {
	    logger.info("computeAllFingerprints Starting");
	    //System.err.println("Inside computeAllFingerprintsUnthreaded");
	    
	    for (File f : base.getFileList()){
		fpList.addAll(fingerprinter.computeFingerprint(f.getPath()));
	    }
	    
	} catch  (SdTextException sdtext) {
            throw sdtext;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }
	
	if (fpList.size() != base.size()){
	    throw new SdTextException("BulkFingerprinter:computeAllFingerprintsUnthreaded:Mismatch in file list and fp list size");
	}
	

	return fpList;
			
    }

}
