package edu.georgetown.gucs.matcher;

import java.util.SortedMap;
import java.util.List;
import java.util.ArrayList;

import edu.georgetown.gucs.tokenizers.*;
import edu.georgetown.gucs.dictionary.*;
import edu.georgetown.gucs.matcher.*;
import edu.georgetown.gucs.utility.*;
import edu.georgetown.gucs.fingerprinter.*;


import org.slf4j.LoggerFactory;

/**
 * This is a class that allows threaded parallel comparison of fingerprints.
 * 
 * @author Clay Shields
 **/

public class BulkMatcher {

    private String matcherName;
    private Integer min_score;

    private static final org.slf4j.Logger logger = LoggerFactory.getLogger(BulkMatcher.class);

    private void init(String matcherType, int score){

	this.matcherName = matcherType;
	this.min_score = score;

    }
    
    /**
     * A constructor for when you want to limit the minimum score
     * @param matcherName
     *           the name of the fingerprinter-derived class to use for fingerprinting
     * @param results
     *      A map to insert the results in
     * @param minScore
     *      The minimum score required to be added to the results
     */
    BulkMatcher(String matcherType, int score){

	init(matcherType, score);
	
    }


    /**
     * A constructor that doesn't require a minimum score; it uses the Global minumum.
     *
     **/
    
    public BulkMatcher(String matcherType) {

	init(matcherType, Global.MIN_SCORE);
    }

     
    /**
     * A method to do multithreaded comparisons between a single fingerprint and a list of other fingerprints
     *
     * @param fing
     *         The single fingerprint to compare
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare the single fingerprint against    
     */
     

    public List<Pair<String, Integer>> compareFingerprintToList(Fingerprint singleFingerprint, List<Fingerprint> fingerprintList) throws SdTextException{

	logger.info("Entry");
	//	System.err.println("Entry into compareFingerprintToList");			  
	
	if (singleFingerprint == null){
	    throw new SdTextException("BulkMatcher:compareFingerprintToList:null fingerprint");
	}

	//	System.err.println("Check fingerprintList");
	if (fingerprintList == null){
	    throw new SdTextException("BulkMatcher:compareFingerprintToList:null fingerprint list");
	}
	
	// Clay - we will fix the threading on this. For the time being, use the unthreaded version
	try{
	    //System.err.println("Calling compareFingerprintToListUnthreaded");
	    logger.info("Calling compareFingerprintToListUnthreaded");

	    List<Pair<String, Integer>> completedMatches =
		Global.pool.invoke(new BulkListMatcherWorker(matcherName,
							     singleFingerprint,
							     fingerprintList,
							     0,
							     fingerprintList.size() -1 ));
	    


	    if (completedMatches.size() != fingerprintList.size()){
		throw new SdTextException("BulkMatcher:compareFingerprintToList:Number of matches differs in size from number of fingerprints to match");
	    }
	    
	    return completedMatches;
	}
	catch (SdTextException e){
	    throw e;
	}
    }


    
    /**
     * A method to do n^2 comparisons among all fingeprints present
     *
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare     
     */
     

    public List<Pair<Pair<String, String>, Integer>> compareAllFingerprints(List<Fingerprint> fingerprints)
	throws SdTextException{
	
	
	return compareAllFingerprintsUnthreaded(fingerprints);
	
    }


    /**
     * A method to do unthreaded comparisons between a single fingerprint and a list of fingerprints
     *
     * @param fing
     *         The single fingerprint to compare
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare the single fingerprint against    
     */ 

    public ArrayList<Pair<String, Integer>> compareFingerprintToListUnthreaded(Fingerprint singleFingerprint, List<Fingerprint> fingerprint) throws SdTextException {

	Integer lowScore = 0;
	Integer okScore = 0;

	// We expect nP scores
	Integer nP = fingerprint.size();
	
        ArrayList<Pair<String, Integer>> arrList = new ArrayList<>();

        try {
            logger.info("FingerprintSize: {}", fingerprint.size());

	    //System.err.println("Builing matcher");
            FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);

	    if (matcher == null){
		throw new SdTextException("BulkMatcher:compareFingerprintToListUnthreaded:ObjectBuilder returned null matcher");
	    }
	    
	    //	    System.err.println("Processing fingerprints");
            for (Fingerprint currentFingerprint : fingerprint) {

                int score = matcher.getScore(singleFingerprint.getFingerprint(), currentFingerprint.getFingerprint());
		if (score >= min_score){
		    arrList.add(new Pair<>(currentFingerprint.getFileName(), score));
		    okScore++;
		}
		else {
		    lowScore++;
		}

            }
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }

	if (nP != (okScore + lowScore)){
	    throw new SdTextException("BulkMatcher:compareFingerprintToListUnthreaded:" + 
				      "Produced fewer fingerprints than expected. Expected: " +
				      nP + " Low: " + lowScore + " Ok: " +
				      okScore);				     
	}
        return arrList;
    }


 /**
     * A method to do unthreaded comparisons between a single fingerprint and a list of fingerprints
     *
     * @param fing
     *         The single fingerprint to compare
     * @param fingerprints
     *           a list of the fingerprint/files pairs to compare the single fingerprint against    
     */ 

    public ArrayList<Pair<Pair<String, String>, Integer>> compareAllFingerprintsUnthreaded(List<Fingerprint> fingerprint)
	throws SdTextException {

	Integer lowScore = 0;
	Integer okScore = 0;

	// We expect (n^2 - n) / 2 scores
	Integer nP = fingerprint.size();
	Integer numberOfScores = ((nP * nP) - nP) / 2;
	
	
        ArrayList<Pair<Pair<String, String>, Integer>> arrList = new ArrayList<>();

        try {

	    // System.err.println("In compare all, list of FP has this many items: " + fingerprint.size());
	    // for (int i = 0; i < fingerprint.size(); i++){
	    // 	System.err.println("     " + fingerprint.get(i).getFileName());
	    // }
	    
            FingerprintMatcher matcher = (FingerprintMatcher) ObjectBuilder.buildObject(matcherName);


	    
	    // Iterate through the files. Files don't need to match themselves, and order
	    // doesn't matter, so it is the lower triangle
	    for (int i = 0; i < fingerprint.size(); i++){
		for (int k = i + 1; k < fingerprint.size(); k++){

		    // Compute the score
		    int score = matcher.getScore(fingerprint.get(i).getFingerprint(), fingerprint.get(k).getFingerprint()); 
		    // Pack up the names
		    Pair<String, String> names = new Pair<String,String>(fingerprint.get(i).getFileName(),
									 fingerprint.get(k).getFileName());
		    // Save the score and the names
		    if (score >= min_score){
			arrList.add(new Pair<>(names, score));
			okScore++;
		    }
		    else {
			lowScore++;
		    }
		}
	    }
        } catch (SdTextException sdtextE) {
            throw sdtextE;
        } catch (Exception e) {
            ExceptionHandler.handleException(e);
        }

	if (numberOfScores != (okScore + lowScore)){
	    throw new SdTextException("BulkMatcher:compareAllFingerprintsUnthreaded:" + 
				      "Produced fewer fingerprints than expected. Expected: " +
				      numberOfScores + " Low: " + lowScore + " Ok: " +
				      okScore);				     
	}
        return arrList;
    }

    
}
