package edu.georgetown.gucs.tokenizers;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.util.List;
import java.util.Vector;
import edu.georgetown.gucs.utility.*;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;

import edu.georgetown.gucs.utility.ExceptionHandler;
import edu.georgetown.gucs.utility.Global;

import org.apache.log4j.LogManager;
import org.apache.log4j.PropertyConfigurator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;


/**
 * Tests tokenizers using a configuration file that contains a list of {@linkplain Tokenizer} objects to use on a
 * document and prints the results of the tokenization
 * 
 * @author Clay Shields
 * 
 */
public class TokenizeFile extends Tokenizer {

  private static final long serialVersionUID = -1504621044845334628L;
  private List<String> tokenizers;
  private String splitter = null;
    private static final Logger logger = LoggerFactory.getLogger(TokenizeFile.class);
    
  /**
   * Constructor that initializes an empty string vector of tokenizers to use
   */
  public TokenizeFile() {

      this.tokenizers = new Vector<String>();
  }

  /**
   * Constructor that sets the tokenizers from a configuration file to use on this document. The first tokenizer in the
   * configuration file must support reading from a file.
   * 
   * @param config_file
   *          the string XML filename that specifies these tokenizers
   */
  public TokenizeFile(String config_file) throws SdTextException{


      this.tokenizers = new Vector<String>();
      try {
	  addTokenizers(config_file);
      } catch (SdTextException e) {
	  throw e;
      }
  }

  /**
   * Adds tokenizers from an XML file to the list of tokenizers to use on this document
   * 
   * @param config_file
   *          the string XML filename that contains the tokenizers
   */
  @SuppressWarnings("unchecked")
  private void addTokenizers(String config_file) throws SdTextException {
      logger.info("Entering addTokenizers");
      try {
	File xml = FileChecker.canRead(config_file,"TokenizeFile","addTokenizers");
      SAXBuilder builder = new SAXBuilder();
      Document doc = builder.build(xml);
      Element dict = doc.getRootElement();
      Element splitterElement;
      for (Element tokenizerElement : (List<Element>) dict.getChildren("tokenizer")) {
        String tokenizer = tokenizerElement.getAttributeValue("name");
        if (this.tokenizers.size() == 0) {
          if (!tokenizer.contains("FileTokenizer")) {
            ExceptionHandler.handleExceptionThrow(new Exception("Dictionary: First tokenizer (" + tokenizer
                + ") must read from a file."));
          } else if ((splitterElement = tokenizerElement.getChild("splitter")) != null) {
            this.splitter = splitterElement.getAttributeValue("name") + " "
                + splitterElement.getAttributeValue("value");
          }
        }

        if (tokenizerElement.getAttributeValue("parameter") != null)
          tokenizer += " " + (tokenizerElement.getAttributeValue("parameter"));
        this.tokenizers.add(tokenizer.trim());
      }
    } catch (SdTextException e) {
	  logger.warn("Caught SdTextException");
	  throw e;
    } catch (Exception e) {
	  logger.warn("Caught Exception");
	ExceptionHandler.handleException(e);
    }
      logger.info("Leaving addTokenizers");
  }

  /**
   * Adds tokenizers from the given list to the list of tokenizers to use on this document
   * 
   * @param tokenizerNames
   *          the list of tokenizer names to use
   */
  public void addTokenizers(List<String> tokenizerNames) {
    this.tokenizers = tokenizerNames;
  }

  /**
   * Splits the given file into tokens and alters or eliminates those tokens based on the vector of tokenizers. Prints
   * each token to standard output. The first tokenizer in the vector of tokenizers must support reading from a file.
   * 
   * @param filename
   *          the name of the file to be tokenized
   */
  public void tokenizeFile(String filename) throws SdTextException {
      try {
	  //System.err.println("In tokenizeFile");
	  FileChecker.canRead(filename, "TokenizeFile", "tokenizeFile");
	  //System.err.println("Making tokenizerList");
	  TokenizerList tokenizerList = new TokenizerList(this.tokenizers);
	  //System.err.println("Setting splitter");
	  tokenizerList.setSplitter(this.splitter);
	  //System.err.println("Tokenizing " + filename);
	  this.tokenVectorMap = tokenizerList.tokenizeFile(filename);
      }
      catch (SdTextException e){
	  throw e;
      }
  }

}
